/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <ortp/ortp.h>
#include "ortp.h"
#include "recv.h"
#include "recv.cmdline.h"
#include "gcc-compat.h"

#define CHUNK_SIZE 65536

extern gint msg_to_buf(mblk_t *, char *, gint);
extern struct gengetopt_args_info conf;

static RtpSession *session;
static int need_header;
static struct timeval start;
static int count, good, c_bad; /* number of (good) chunks / consecutive bad chunks */
static guint32 timestamp;
static struct receiver *self; /* pointer to our own receiver struct */

static int handle_other_packet(mblk_t *mp)
{
	char buf[CHUNK_SIZE];
	int ret = -1, code, len = msg_to_buf(mp, buf, CHUNK_SIZE);

	if (len <= 0)
		return -E_MSG_TO_BUF;
	code = buf[0];
	PARA_INFO_LOG("rcvd ctrl pkg #%i (%d bytes)\n", code, len);
	switch (code) {
	case ORTP_EOF:
		ret = 0;
		goto out;
		break;
	case ORTP_BOF:
		need_header = 0;
		gettimeofday(&start, NULL);
		c_bad = 1; /* force default timeout */
		break;
	case ORTP_EMPTY:
		if (!start.tv_sec)
			gettimeofday(&start, NULL);
		count++;
		c_bad = 0;
		break;
	case ORTP_HEADER:
		if (!need_header)
			break; /* just ignore it */
		need_header = 0;
		if (!self->buffer) /* first chunk that gets sent */
			self->buffer = para_malloc(CHUNK_SIZE);
		memmove(self->buffer, buf + 1, len - 1);
		self->loaded = len - 1;
		PARA_NOTICE_LOG("have header: %d bytes\n", len - 1);
		c_bad = 1; /* force default timeout */
		gettimeofday(&start, NULL);
		break;
	default:
		PARA_WARNING_LOG("%s", "rcvd unknown control pkg\n");
		goto out;
	}
	ret = 1;
out:
	freemsg(mp);
	return ret;
}

static int ortp_setup_select_parms(__unused fd_set *rfds, __unused fd_set *wfds, struct timeval *timeout)
{
	struct timeval ret = {0, 5000}, tmp, now, next_chunk;

	if (!start.tv_sec) {/* not yet started */
		timestamp = 0;
		goto out;
	}
	gettimeofday(&now, NULL);
	if (!good || c_bad) { /* previous read failed, use current ts */
		tv_diff(&now, &start, &tmp);
		timestamp = rtp_session_time_to_ts(session, tv2ms(&tmp));
		goto out; /* and default timeout */
	}
	tv_scale(count, &self->chunk_tv, &tmp);
	timestamp = rtp_session_time_to_ts(session, tv2ms(&tmp));
	tv_add(&tmp, &start, &next_chunk);
	if (tv_diff(&next_chunk, &now, &ret) < 0) {
		/* overdue, sleep only 1ms */
		ret.tv_sec = 0;
		ret.tv_usec = 1000;
	}
out:
	if (tv_diff(timeout, &ret, NULL) > 0)
		*timeout = ret;
	return -1; /* we did not modify rfds */
}


int ortp_handle_fds(int select_ret, __unused fd_set *rfds, __unused fd_set *wfds)
{
	mblk_t *mp = rtp_session_recvm_with_ts(session, timestamp);
	int ret;

	if (select_ret) /* we're only interested in timeouts */
		return 1;
	if (!mp) {
		c_bad++;
		if (!(c_bad % 100))
			PARA_DEBUG_LOG("did not receive  buffer, ts=%u, count = %d, good = %d, c_bad = %d\n",
				timestamp, count, good, c_bad);
		if (good && c_bad > 500) {
			PARA_ERROR_LOG("%d consecutive bad chunks, giving up\n", c_bad);
			return -E_READ;
		}
		goto out;
	}
	PARA_DEBUG_LOG("have it ts = %d, good = %d\n", timestamp, good);
	if (rtp_session_get_recv_payload_type(session) == PAYLOAD_OTHER)
		return handle_other_packet(mp);
	c_bad = 0;
	count++;
	if (!start.tv_sec)
		gettimeofday(&start, NULL);
	/* okay, we have a chunk of data */
	if (need_header) {
		/*
		 * but we can not use it because we need the audio file header and
		 * don't have it yet. Just ignore the data
		 */
		return 1;
	}
	good++;
	if (self->loaded >= CHUNK_SIZE)
		return -E_OVERRUN;
	if (!self->buffer) /* first chunk that gets sent */
		self->buffer = para_malloc(CHUNK_SIZE);
	ret = msg_to_buf(mp, self->buffer + self->loaded,
		CHUNK_SIZE - self->loaded);
	if (ret <= 0)
		return -E_MSG_TO_BUF;
	self->loaded += ret;
out:
	return 1;
}

static void ortp_shutdown(void)
{
//	ortp_global_stats_display();
//	ortp_exit();
	if (session)
		rtp_session_destroy(session);
	free(self->buffer);
}

int ortp_recv_init(struct receiver *r)
{
	ortp_init();
	ortp_set_debug_file("oRTP", NULL);
	session = rtp_session_new(RTP_SESSION_RECVONLY);
	PARA_NOTICE_LOG("receiving from %s:%d\n", conf.ortp_host_arg,
		conf.ortp_port_arg);
	rtp_session_set_local_addr(session, conf.ortp_host_arg, conf.ortp_port_arg);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
//	rtp_session_set_jitter_compensation(session, 100);
	good = 0;
	c_bad = 0;
	start.tv_sec = 0;
	need_header = conf.header_given? 1 : 0;

	r->shutdown = ortp_shutdown;
	r->setup_select_parms = ortp_setup_select_parms;
	r->handle_fds = ortp_handle_fds;
	r->chunk_tv.tv_sec = 0;
	r->chunk_tv.tv_usec = conf.chunk_time_arg;
	r->loaded = 0;
	r->buffer = NULL;
	self = r;
	return 1;
}
